#![doc = include_str!("../README.md")]

#[cfg(feature = "macros")]
pub mod macros;
pub mod query;
pub mod result;
#[cfg(feature = "utils")]
pub mod utils;

use std::fmt::Display;

use prometheus_http_query::Client;
use query::{ops::Operable, IntoQuery};
use reqwest::header::HeaderValue;
use result::{IntoResult, QryResult, RawResult};
use seal::Sealed;
use url::Url;

#[cfg(feature = "multi")]
use tokio::task::{JoinError, JoinSet};

/// The primary way to send queries and get deserialized responses from the metrics server
///
/// The default instance uses no authentication, initializes it's own reqwest client and the
/// server url is set to "http://127.0.0.1:9090" and sends requests using the GET HTTP method.
#[derive(Default, Clone)]
pub struct QueryManager {
    auth: Auth,
    client: Client,
    timeout: Option<i64>,
    method: Method,
}

impl QueryManager {
    /// Build a new QueryManager using a valid base URL (without Prometheus API path)
    pub fn new(url: Url) -> Self {
        QueryManager {
            #[allow(clippy::unwrap_used)] // url is already validated
            client: Client::try_from(url.as_str()).expect("unexpected failure"),
            ..Default::default()
        }
    }

    /// Build a new QueryManager using a valid base URL (without Prometheus API path) and an existing reqwest Client
    pub fn from_client(url: Url, client: reqwest::Client) -> Self {
        QueryManager {
            auth: Default::default(),
            #[allow(clippy::unwrap_used)] // url is already validated
            client: Client::from(client, url.as_str()).expect("unexpected failure"),
            ..Default::default()
        }
    }

    /// Set authentication method (no auth is used by default)
    ///
    /// Currently only supports bearer tokens
    pub fn auth(&mut self, auth: Auth) -> &mut Self {
        self.auth = auth;
        self
    }

    /// Set timeout for Prometheus queries
    pub fn timeout(&mut self, timeout: i64) -> &mut Self {
        self.timeout.replace(timeout);
        self
    }

    /// Set HTTP method to use to send queries to the server (default: `Get`)
    pub fn method(&mut self, method: Method) -> &mut Self {
        self.method = method;
        self
    }

    /// Send a query to the Prometheus server and retreive deserialized data
    ///
    /// This method accepts any type that implements the `IntoQuery` trait.
    /// The trait is implemented for `&str`, `String`, `&String`, and __mquery's__
    /// internal types such as [`query::Metric`], [`query::Scalar`] and others that result
    /// from the query builder API.
    pub async fn query<Q: IntoQuery>(&self, query: Q) -> QryResult {
        let mut builder = self.client.query(query.into_query().as_ref());
        if let Some((name, val)) = self.auth.get_header() {
            builder = builder.header(name, val);
        }
        if let Some(timeout) = self.timeout.as_ref() {
            builder = builder.timeout(*timeout);
        }
        match self.method {
            Method::Get => builder.get().await.into_result(),
            Method::Post => builder.post().await.into_result(),
        }
    }

    /// Send a query to the Prometheus server and retreive raw response ([`reqwest::Response`])
    ///
    /// This method accepts any type that implements the `IntoQuery` trait.
    /// The trait is implemented for `&str`, `String`, `&String`, and __mquery's__
    /// internal types such as [`query::Metric`], [`query::Scalar`] and others that result
    /// from the query builder API.
    pub async fn query_raw<Q: IntoQuery>(&self, query: Q) -> RawResult {
        let mut builder = self.client.query(query.into_query().as_ref());
        if let Some((name, val)) = self.auth.get_header() {
            builder = builder.header(name, val);
        }
        if let Some(timeout) = self.timeout.as_ref() {
            builder = builder.timeout(*timeout);
        }
        match self.method {
            Method::Get => builder.get_raw().await.into_result(),
            Method::Post => builder.post_raw().await.into_result(),
        }
    }

    /// Send multiple queries to the Prometheus server and get a Vec of deserialized results
    ///
    /// This method accepts an iterator of any type that implements the `IntoQuery` trait.
    /// The trait is implemented for `&str`, `String`, `&String`, and __mquery's__
    /// internal types such as [`query::Metric`], [`query::Scalar`] and others that result
    /// from the query builder API.
    ///
    /// If there are different types of queries being used, you can turn them all into Strings
    /// using the `to_string` method before passing them to the `queries` parameter.
    #[cfg(feature = "multi")]
    pub async fn query_multi<Q: IntoQuery + Send + Sync + 'static>(
        &self,
        queries: impl IntoIterator<Item = Q>,
    ) -> Result<Vec<QryResult>, JoinError> {
        let Self {
            auth,
            client,
            timeout,
            method,
        } = self;

        let mut set = JoinSet::new();
        for q in queries {
            set.spawn(query(q, auth.clone(), *timeout, client.clone(), *method));
        }
        let mut res = Vec::with_capacity(set.len());
        while let Some(qr) = set.join_next().await {
            res.push(qr?);
        }
        Ok(res)
    }

    /// Send a ranged query to the Prometheus server and retreived deserialized data
    ///
    /// This method accepts any type that implements the `IntoQuery` trait.
    /// The trait is implemented for `&str`, `String`, `&String`, and __mquery's__
    /// internal types such as [`query::Metric`], [`query::Scalar`] and others that result
    /// from the query builder API.
    pub async fn query_range<Q: IntoQuery>(
        &self,
        query: Q,
        start: i64,
        end: i64,
        step: f64,
    ) -> QryResult {
        let mut builder = self
            .client
            .query_range(query.into_query().as_ref(), start, end, step);
        if let Some((name, val)) = self.auth.get_header() {
            builder = builder.header(name, val);
        }
        if let Some(timeout) = self.timeout.as_ref() {
            builder = builder.timeout(*timeout);
        }
        match self.method {
            Method::Get => builder.get().await.into_result(),
            Method::Post => builder.post().await.into_result(),
        }
    }
}

// TODO: Add support for basic auth as well
/// Authentication method
///
/// Used in [`QueryManager::auth`] to set the authentication type
#[derive(Default, Clone, Debug)]
pub enum Auth {
    #[default]
    None,
    Bearer(String),
}

impl Auth {
    const AUTH_HEADER: &'static str = "Authorization";

    fn get_header(&self) -> Option<(&'static str, HeaderValue)> {
        match self {
            Auth::None => None,
            Auth::Bearer(token) => Some((
                Self::AUTH_HEADER,
                #[allow(clippy::unwrap_used)] // building from valid chars
                format!("Bearer {}", token)
                    .parse()
                    .expect("unexpected failure"),
            )),
        }
    }
}

/// HTTP method to use to send queries
///
/// This is used by the [`QueryManager::method`] method to set the HTTP
/// method used to send queries to the server.
#[derive(Clone, Copy, Debug, Default)]
pub enum Method {
    #[default]
    Get,
    Post,
}

// Private trait used to seal other traits to disallow users from implementing
// them directly
mod seal {
    pub trait Sealed {}
}

/// Signifies that a query has been validated through some of the validation methods
/// offered by the library
pub struct ValidatedQuery<T: AsRef<str>>(pub T);

impl<T: AsRef<str>> IntoQuery for ValidatedQuery<T> {
    type Target = T;

    fn into_query(self) -> Self::Target {
        self.0
    }
}

impl<T: AsRef<str>> Display for ValidatedQuery<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.as_ref().fmt(f)
    }
}

impl<T: AsRef<str>> Operable for ValidatedQuery<T> {}
impl<T: AsRef<str>> Sealed for ValidatedQuery<T> {}

#[cfg(feature = "multi")]
async fn query<Q: IntoQuery>(
    query: Q,
    auth: Auth,
    timeout: Option<i64>,
    client: Client,
    method: Method,
) -> QryResult {
    let mut builder = client.query(query.into_query().as_ref());
    if let Some((name, val)) = auth.get_header() {
        builder = builder.header(name, val);
    }
    if let Some(timeout) = timeout.as_ref() {
        builder = builder.timeout(*timeout);
    }
    match method {
        Method::Get => builder.get().await.into_result(),
        Method::Post => builder.post().await.into_result(),
    }
}
